from __future__ import print_function
from optparse import OptionParser
from classifyUtil.morphologicalCovariance import erodeByVector,morphologicalCovariance
import itertools
import os
import PIL.Image as Image
import re
import scipy as sp
import scipy.linalg as la

parser = OptionParser(usage="Usage: %prog [options] images")
parser.add_option("-s", "--space", type=str, default="sphere")
parser.add_option("--prefix", type=str, default="Outex_TC_00013")
parser.add_option("--maxn", type=int, default=150)
parser.add_option('--project', action="store_true", default=False, help="Project back to original colour space before computing feature.")
(options,args) = parser.parse_args()

try:
    # The __import__ statement is a little weird, when importing a submodule, it only returns the submodule (instead of the top-level module) if fromlist is given (and non-empty).
    space = __import__("spaces.%s" % options.space, fromlist=[''])
except ImportError:
    parser.error("Unsupported space: %s." % options.space)

spaceName = options.space
projectBack = options.project
prefix = options.prefix

if projectBack:
    spaceName = "%s-projected" % spaceName

if space.n:
    maxn = space.n
else:
    maxn = options.maxn

dirName = "%s/features/%s" % (prefix,spaceName)

try:
    os.makedirs(dirName)
except OSError:
    # Should be okay, or otherwise we will fail later on anyway.
    # The reason to do this instead of checking whether the directory exists, is that the latter can lead to a race condition.
    pass

# Coordinates are "inverted" (y,x) instead of (x,y)
vectors0 = [(0,d) for d in range(1,50,2)]
vectors45 = [(d,d) for d in range(1,50,2)]
vectors90 = [(d,0) for d in range(1,50,2)]
vectors135 = [(d,-d) for d in range(1,50,2)]
vectors = vectors0 + vectors45 + vectors90 + vectors135
assert(len(vectors)==100)

# Get list of directions (we only support deterministic frames)
dgen = space.sample()
D = sp.array(list(itertools.islice(dgen,maxn)))
offsets = map(lambda d:-255.0*min([0,sp.dot(d,[0,0,1]),sp.dot(d,[0,1,0]),sp.dot(d,[0,1,1]),sp.dot(d,[1,0,0]),sp.dot(d,[1,0,1]),sp.dot(d,[1,1,0]),sp.dot(d,[1,1,1])]), D)

if projectBack:
    DOrg = D
    D = sp.array([[1,0,0],[0,1,0],[0,0,1]])
    offsets = [0,0,0]
    invD = la.pinv(D).T

# Process files
for filename in args:
    basename = re.match('.*/([^/]+)\.\w+$', filename).group(1)
    print("Processing %s" % basename)

    fileNameBase = "%s/%s" % (dirName,basename)

    I = sp.array(Image.open(filename),dtype="float")
    Isums = I.sum(axis=0).sum(axis=0)
    
    # Compute feature vectors
    
    covs = sp.zeros((len(D),len(vectors)))
    if projectBack:
        # This is equivalent to eroding G in the frame, projecting back and then summing.
        #   All channels are eroded independently, so we can compute the contribution of a channel
        #   to the final erosion by computing the channel, eroding it and projecting back just that channel.
        #   This we do for all channels and all erosion vectors. Finally we divide by the normalisation constant(s).
        for d,invd in zip(DOrg,invD):
            G = sp.tensordot(I,d,axes=(2,0))
            for vi,vec in enumerate(vectors):
                covs[:,vi] += erodeByVector(G,vec).sum()*invd
        covs /= Isums[:,sp.newaxis]
    else:
        for di,d,offset in zip(range(len(D)),D,offsets):
            G = sp.tensordot(I,d,axes=(2,0)) + offset
            for vi,vec in enumerate(vectors):
                covs[di,vi] = morphologicalCovariance(G,vec)
    
    # Write out feature vectors
    
    fileName = "%s.csv" % fileNameBase
    with open(fileName,"wb") as featureFile:
        featureFile.write('"dx","dy","dz"')
        for vec in vectors:
            featureFile.write(',"%s"' % str(vec))
        featureFile.write('\n')
        for di,d in enumerate(D):
            featureFile.write("%.8f,%.8f,%.8f" % tuple(d))
            for vi in xrange(len(vectors)):
                featureFile.write(",%.8g" % covs[di,vi])
            featureFile.write("\n")

